"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.convertNodeSchema = exports.convertAllowedTypes = exports.convertField = exports.getStoredSchema = exports.getFlexSchema = exports.toStoredSchema = exports.toFlexSchema = void 0;
const internal_1 = require("@fluidframework/core-utils/internal");
const internal_2 = require("@fluidframework/telemetry-utils/internal");
const index_js_1 = require("../core/index.js");
const index_js_2 = require("../feature-libraries/index.js");
// TODO: once flex schema is gone, this code can move into simple-tree
// eslint-disable-next-line import/no-internal-modules
const index_js_3 = require("../feature-libraries/typed-schema/index.js");
const index_js_4 = require("../util/index.js");
const index_js_5 = require("./core/index.js");
const schemaTypes_js_1 = require("./schemaTypes.js");
/**
 * Generate a {@link FlexTreeSchema} with `root` as the root field.
 *
 * This also has the side effect of populating the cached view schema on the class-based schema.
 */
function toFlexSchema(root) {
    const schemaMap = new Map();
    const field = convertField(schemaMap, root);
    const nodeSchema = new Map((0, index_js_4.mapIterable)(schemaMap, ([key, value]) => {
        const schema = value.toFlex();
        const classSchema = (0, index_js_5.tryGetSimpleNodeSchema)(schema);
        if (classSchema === undefined) {
            (0, internal_1.assert)((0, index_js_2.schemaIsLeaf)(schema), 0x83e /* invalid leaf */);
        }
        else {
            (0, internal_1.assert)((0, index_js_5.cachedFlexSchemaFromClassSchema)(classSchema) === schema, 0x83f /* mismatched schema */);
        }
        return [key, schema];
    }));
    const typed = {
        nodeSchema,
        adapters: {},
        rootFieldSchema: field,
        policy: index_js_2.defaultSchemaPolicy,
    };
    return typed;
}
exports.toFlexSchema = toFlexSchema;
/**
 * Converts a {@link ImplicitFieldSchema} into a {@link TreeStoredSchema}.
 */
function toStoredSchema(root) {
    const flex = toFlexSchema(root);
    return {
        rootFieldSchema: flex.rootFieldSchema.stored,
        ...(0, index_js_2.intoStoredSchemaCollection)(flex),
    };
}
exports.toStoredSchema = toStoredSchema;
/**
 * Return a flex schema for the provided class schema.
 *
 * This also has the side effect of populating the cached view schema on the class based schema.
 */
function getFlexSchema(root) {
    const treeSchema = toFlexSchema(root);
    return treeSchema.rootFieldSchema.monomorphicChildType ?? (0, index_js_4.fail)("root should be monomorphic");
}
exports.getFlexSchema = getFlexSchema;
/**
 * Return a stored schema for the provided class schema.
 *
 * This also has the side effect of populating the cached view schema on the class based schema.
 */
function getStoredSchema(root) {
    return getFlexSchema(root).stored;
}
exports.getStoredSchema = getStoredSchema;
/**
 * Normalizes an {@link ImplicitFieldSchema} into a {@link TreeFieldSchema}.
 */
function convertField(schemaMap, schema) {
    let kind;
    let types;
    if (schema instanceof schemaTypes_js_1.FieldSchema) {
        kind = convertFieldKind.get(schema.kind) ?? (0, index_js_4.fail)("Invalid field kind");
        types = schema.allowedTypes;
    }
    else {
        kind = index_js_2.FieldKinds.required;
        types = schema;
    }
    const allowedTypes = convertAllowedTypes(schemaMap, types);
    return index_js_2.FlexFieldSchema.create(kind, allowedTypes);
}
exports.convertField = convertField;
const convertFieldKind = new Map([
    [schemaTypes_js_1.FieldKind.Optional, index_js_2.FieldKinds.optional],
    [schemaTypes_js_1.FieldKind.Required, index_js_2.FieldKinds.required],
    [schemaTypes_js_1.FieldKind.Identifier, index_js_2.FieldKinds.identifier],
]);
/**
 * Normalizes an {@link ImplicitAllowedTypes} into an {@link AllowedTypes}.
 */
function convertAllowedTypes(schemaMap, schema) {
    if ((0, index_js_4.isReadonlyArray)(schema)) {
        return (0, index_js_3.normalizeFlexListEager)(schema).map((item) => convertNodeSchema(schemaMap, item));
    }
    return [convertNodeSchema(schemaMap, schema)];
}
exports.convertAllowedTypes = convertAllowedTypes;
const builder = { name: "simple schema" };
/**
 * Converts a {@link TreeNodeSchema} into a {@link FlexTreeNodeSchema}.
 * Ensures all types reachable from `schema` are included in `schemaMap`.
 *
 * Return value (and entries in map) are lazy to allow recursive types to work.
 * This laziness does NOT extend to adding entries to `schemaMap`:
 * all referenced types are added to it before this function returns.
 */
function convertNodeSchema(schemaMap, schema) {
    const fromMap = schemaMap.get((0, index_js_4.brand)(schema.identifier));
    if (fromMap !== undefined) {
        if (fromMap.original !== schema) {
            // Use JSON.stringify to quote and escape identifier string.
            throw new internal_2.UsageError(`Multiple schema encountered with the identifier ${JSON.stringify(schema.identifier)}. Remove or rename them to avoid the collision.`);
        }
        return fromMap.toFlex;
    }
    const toFlex = () => {
        let out;
        const kind = schema.kind;
        switch (kind) {
            case index_js_5.NodeKind.Leaf: {
                const cached = (0, index_js_5.cachedFlexSchemaFromClassSchema)(schema) ?? (0, index_js_4.fail)("leaf schema should be pre-cached");
                (0, internal_1.assert)((0, index_js_2.schemaIsLeaf)(cached), 0x840 /* expected leaf */);
                return cached;
            }
            case index_js_5.NodeKind.Map: {
                const fieldInfo = schema.info;
                const field = index_js_2.FlexFieldSchema.create(index_js_2.FieldKinds.optional, convertAllowedTypes(schemaMap, fieldInfo));
                // Lookup of cached schema is done here instead of before since walking the schema recursively to populate schemaMap is still required.
                const cached = (0, index_js_5.cachedFlexSchemaFromClassSchema)(schema);
                out =
                    cached ??
                        index_js_2.FlexMapNodeSchema.create(builder, (0, index_js_4.brand)(schema.identifier), field);
                break;
            }
            case index_js_5.NodeKind.Array: {
                const fieldInfo = schema.info;
                const field = index_js_2.FlexFieldSchema.create(index_js_2.FieldKinds.sequence, convertAllowedTypes(schemaMap, fieldInfo));
                const cached = (0, index_js_5.cachedFlexSchemaFromClassSchema)(schema);
                out =
                    cached ??
                        index_js_2.FlexObjectNodeSchema.create(builder, (0, index_js_4.brand)(schema.identifier), {
                            [index_js_1.EmptyKey]: field,
                        });
                break;
            }
            case index_js_5.NodeKind.Object: {
                const info = schema.info;
                const fields = Object.create(null);
                for (const [propertyKey, implicitFieldSchema] of Object.entries(info)) {
                    // If a `stored key` was provided, use it as the key in the flex schema.
                    // Otherwise, use the property key.
                    const flexKey = (0, schemaTypes_js_1.getStoredKey)(propertyKey, implicitFieldSchema);
                    // This code has to be careful to avoid assigning to __proto__ or similar built-in fields.
                    Object.defineProperty(fields, flexKey, {
                        enumerable: true,
                        configurable: false,
                        writable: false,
                        value: convertField(schemaMap, implicitFieldSchema),
                    });
                }
                const cached = (0, index_js_5.cachedFlexSchemaFromClassSchema)(schema);
                out = cached ?? index_js_2.FlexObjectNodeSchema.create(builder, (0, index_js_4.brand)(schema.identifier), fields);
                break;
            }
            default:
                (0, internal_1.unreachableCase)(kind);
        }
        (0, internal_1.assert)(out instanceof index_js_2.TreeNodeSchemaBase, 0x841 /* invalid schema produced */);
        {
            const cached = (0, index_js_5.cachedFlexSchemaFromClassSchema)(schema);
            if (cached !== undefined) {
                (0, internal_1.assert)((0, index_js_5.cachedFlexSchemaFromClassSchema)(schema) === out, 0x842 /* incorrect flexSchemaSymbol */);
            }
            else {
                (0, index_js_5.setFlexSchemaFromClassSchema)(schema, out);
            }
        }
        return out;
    };
    schemaMap.set((0, index_js_4.brand)(schema.identifier), { original: schema, toFlex });
    return toFlex;
}
exports.convertNodeSchema = convertNodeSchema;
//# sourceMappingURL=toFlexSchema.js.map